"""
    This package contains interfaces for heuristic functions that are used in variants of the 'SeqClu' algorithm.
"""

from .prototype_value import IPrototypeValue


__all__ = ["IPrototypeValue"]
