"""
    This package contains implementations of heuristic functions that are used in variants of the 'SeqClu' algorithm.
"""

from .linear_prototype_value import LinearPrototypeValue


__all__ = ["LinearPrototypeValue"]
