"""
    This module contains classes that bundle several approaches to visualize the results of the variations of
    the 'SeqClu' algorithm that are contained in the package.
"""
from .visualizer import Visualizer

__all__ = ["Visualizer"]
