"""
    This package contains three data sets, namely:
    1. GesturePebbleZ1 (http://www.timeseriesclassification.com/description.php?Dataset=GesturePebbleZ1)
    2. UJI Pen Characters (https://archive.ics.uci.edu/ml/datasets/UJI+Pen+Characters)
    3. PLAID (http://www.timeseriesclassification.com/description.php?Dataset=PLAID)

    Additionally, the package contains the original source code provided at the start of the project as a Jupyter
    Notebook.
"""

__all__ = []
